from pathlib import Path
from typing import Literal

from confz import BaseConfig, EnvSource


class OperationMode(BaseConfig):  # type: ignore[misc]
    environment: Literal["development", "production"] = "development"

    CONFIG_SOURCES = EnvSource(allow=["ENVIRONMENT"])


class ServiceConfig(BaseConfig):  # type: ignore[misc]
    config_dir: Path = Path("config")
    web_port: int = 8001
    rpc_port: int = 18871

    CONFIG_SOURCES = EnvSource(prefix="SERVICE_")


class WebServerConfig(BaseConfig):  # type: ignore[misc]
    generate_web_settings: bool = False

    CONFIG_SOURCES = EnvSource(allow=["GENERATE_WEB_SETTINGS"])



from __future__ import annotations

import asyncio
import inspect
import logging
from typing import TYPE_CHECKING, Any, TypedDict

from pydase.data_service.abstract_data_service import AbstractDataService
from pydase.utils.helpers import get_class_and_instance_attributes

if TYPE_CHECKING:
    from collections.abc import Callable

    from .data_service import DataService

logger = logging.getLogger(__name__)


class TaskDict(TypedDict):
    task: asyncio.Task[None]
    kwargs: dict[str, Any]


class TaskManager:
    """
    The TaskManager class is a utility designed to manage asynchronous tasks. It
    provides functionality for starting, stopping, and tracking these tasks. The class
    is primarily used by the DataService class to manage its tasks.

    A task in TaskManager is any asynchronous function. To add a task, you simply need
    to define an async function within your class that extends TaskManager. For example:

    ```python
    class MyService(DataService):
        async def my_task(self):
            # Your task implementation here
            pass
    ```

    With the above definition, TaskManager automatically creates `start_my_task` and
    `stop_my_task` methods that can be used to control the task.

    TaskManager also supports auto-starting tasks. If there are tasks that should start
    running as soon as an instance of your class is created, you can define them in
    `self._autostart_tasks` in your class constructor (__init__ method). Here's how:

    ```python
    class MyService(DataService):
        def __init__(self):
            self._autostart_tasks = {
                "my_task": (*args)  # Replace with actual arguments
            }
            self.wait_time = 1
            super().__init__()

        async def my_task(self, *args):
            while True:
                # Your task implementation here
                await asyncio.sleep(self.wait_time)
    ```

    In the above example, `my_task` will start running as soon as
    `_start_autostart_tasks` is called which is done when the DataService instance is
    passed to the `pydase.Server` class.

    The responsibilities of the TaskManager class are:

    - Track all running tasks: Keeps track of all the tasks that are currently running.
    This allows for monitoring of task statuses and for making sure tasks do not
    overlap.
    - Provide the ability to start and stop tasks: Automatically creates methods to
    start and stop each task.
    - Emit notifications when the status of a task changes: Has a built-in mechanism for
    emitting notifications when a task starts or stops. This is used to update the user
    interfaces, but can also be used to write logs, etc.
    """

    def __init__(self, service: DataService) -> None:
        self.service = service

        self.tasks: dict[str, TaskDict] = {}
        """A dictionary to keep track of running tasks. The keys are the names of the
        tasks and the values are TaskDict instances which include the task itself and
        its kwargs.
        """

        self._set_start_and_stop_for_async_methods()

    @property
    def _loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    def _set_start_and_stop_for_async_methods(self) -> None:
        # inspect the methods of the class
        for name, method in inspect.getmembers(
            self.service, predicate=inspect.iscoroutinefunction
        ):
            # create start and stop methods for each coroutine
            setattr(self.service, f"start_{name}", self._make_start_task(name, method))
            setattr(self.service, f"stop_{name}", self._make_stop_task(name))

    def _initiate_task_startup(self) -> None:
        if self.service._autostart_tasks is not None:
            for service_name, args in self.service._autostart_tasks.items():
                start_method = getattr(self.service, f"start_{service_name}", None)
                if start_method is not None and callable(start_method):
                    start_method(*args)
                else:
                    logger.warning(
                        "No start method found for service '%s'", service_name
                    )

    def start_autostart_tasks(self) -> None:
        self._initiate_task_startup()
        attrs = get_class_and_instance_attributes(self.service)

        for attr_value in attrs.values():
            if isinstance(attr_value, AbstractDataService):
                attr_value._task_manager.start_autostart_tasks()
            elif isinstance(attr_value, list):
                for item in attr_value:
                    if isinstance(item, AbstractDataService):
                        item._task_manager.start_autostart_tasks()

    def _make_stop_task(self, name: str) -> Callable[..., Any]:
        """
        Factory function to create a 'stop_task' function for a running task.

        The generated function cancels the associated asyncio task using 'name' for
        identification, ensuring proper cleanup. Avoids closure and late binding issues.

        Args:
            name (str): The name of the coroutine task, used for its identification.
        """

        def stop_task() -> None:
            # cancel the task
            task = self.tasks.get(name, None)
            if task is not None:
                self._loop.call_soon_threadsafe(task["task"].cancel)

        return stop_task

    def _make_start_task(
        self, name: str, method: Callable[..., Any]
    ) -> Callable[..., Any]:
        """
        Factory function to create a 'start_task' function for a coroutine.

        The generated function starts the coroutine as an asyncio task, handling
        registration and monitoring.
        It uses 'name' and 'method' to avoid the closure and late binding issue.

        Args:
            name (str): The name of the coroutine, used for task management.
            method (callable): The coroutine to be turned into an asyncio task.
        """

        def start_task(*args: Any, **kwargs: Any) -> None:
            def task_done_callback(task: asyncio.Task[None], name: str) -> None:
                """Handles tasks that have finished.

                Removes a task from the tasks dictionary, calls the defined
                callbacks, and logs and re-raises exceptions."""

                # removing the finished task from the tasks i
                self.tasks.pop(name, None)

                # emit the notification that the task was stopped
                self.service._notify_changed(name, None)

                exception = task.exception()
                if exception is not None:
                    # Handle the exception, or you can re-raise it.
                    logger.error(
                        "Task '%s' encountered an exception: %s: %s",
                        name,
                        type(exception).__name__,
                        exception,
                    )
                    raise exception

            async def task(*args: Any, **kwargs: Any) -> None:
                try:
                    await method(*args, **kwargs)
                except asyncio.CancelledError:
                    logger.info("Task '%s' was cancelled", name)

            if not self.tasks.get(name):
                # Get the signature of the coroutine method to start
                sig = inspect.signature(method)

                # Create a list of the parameter names from the method signature.
                parameter_names = list(sig.parameters.keys())

                # Extend the list of positional arguments with None values to match
                # the length of the parameter names list. This is done to ensure
                # that zip can pair each parameter name with a corresponding value.
                args_padded = list(args) + [None] * (len(parameter_names) - len(args))

                # Create a dictionary of keyword arguments by pairing the parameter
                # names with the values in 'args_padded'. Then merge this dictionary
                # with the 'kwargs' dictionary. If a parameter is specified in both
                # 'args_padded' and 'kwargs', the value from 'kwargs' is used.
                kwargs_updated = {
                    **dict(zip(parameter_names, args_padded, strict=True)),
                    **kwargs,
                }

                # creating the task and adding the task_done_callback which checks
                # if an exception has occured during the task execution
                task_object = self._loop.create_task(task(*args, **kwargs))
                task_object.add_done_callback(
                    lambda task: task_done_callback(task, name)
                )

                # Store the task and its arguments in the '__tasks' dictionary. The
                # key is the name of the method, and the value is a dictionary
                # containing the task object and the updated keyword arguments.
                self.tasks[name] = {
                    "task": task_object,
                    "kwargs": kwargs_updated,
                }

                # emit the notification that the task was started
                self.service._notify_changed(name, kwargs_updated)
            else:
                logger.error("Task '%s' is already running!", name)

        return start_task



import logging
from typing import Any

from pydase.observer_pattern.observable.observable_object import ObservableObject
from pydase.utils.helpers import is_property_attribute

logger = logging.getLogger(__name__)


class Observable(ObservableObject):
    def __init__(self) -> None:
        super().__init__()
        class_attrs = {
            k: type(self).__dict__[k]
            for k in set(type(self).__dict__)
            - set(Observable.__dict__)
            - set(self.__dict__)
        }
        for name, value in class_attrs.items():
            if isinstance(value, property) or callable(value):
                continue
            self.__dict__[name] = self._initialise_new_objects(name, value)

    def __setattr__(self, name: str, value: Any) -> None:
        if not hasattr(self, "_observers") and name != "_observers":
            logger.warning(
                "Ensure that super().__init__() is called at the start of the '%s' "
                "constructor! Failing to do so may lead to unexpected behavior.",
                type(self).__name__,
            )
            self._observers = {}

        value = self._handle_observable_setattr(name, value)

        super().__setattr__(name, value)

        self._notify_changed(name, value)

    def __getattribute__(self, name: str) -> Any:
        if is_property_attribute(self, name):
            self._notify_change_start(name)

        value = super().__getattribute__(name)

        if is_property_attribute(self, name):
            self._notify_changed(name, value)

        return value

    def _handle_observable_setattr(self, name: str, value: Any) -> Any:
        if name == "_observers":
            return value

        self._remove_observer_if_observable(name)
        value = self._initialise_new_objects(name, value)
        self._notify_change_start(name)
        return value

    def _remove_observer_if_observable(self, name: str) -> None:
        if not is_property_attribute(self, name):
            current_value = getattr(self, name, None)

            if isinstance(current_value, ObservableObject):
                current_value._remove_observer(self, name)

    def _construct_extended_attr_path(
        self, observer_attr_name: str, instance_attr_name: str
    ) -> str:
        if observer_attr_name != "":
            return f"{observer_attr_name}.{instance_attr_name}"
        return instance_attr_name



import logging
from typing import TYPE_CHECKING, Any

from pydase.utils.serializer import (
    SerializationPathError,
    SerializationValueError,
    get_nested_dict_by_path,
    set_nested_value_by_path,
)

if TYPE_CHECKING:
    from pydase import DataService

logger = logging.getLogger(__name__)


class DataServiceCache:
    def __init__(self, service: "DataService") -> None:
        self._cache: dict[str, Any] = {}
        self.service = service
        self._initialize_cache()

    @property
    def cache(self) -> dict[str, Any]:
        return self._cache

    def _initialize_cache(self) -> None:
        """Initializes the cache and sets up the callback."""
        logger.debug("Initializing cache.")
        self._cache = self.service.serialize()

    def update_cache(self, full_access_path: str, value: Any) -> None:
        set_nested_value_by_path(self._cache, full_access_path, value)

    def get_value_dict_from_cache(self, full_access_path: str) -> dict[str, Any]:
        try:
            return get_nested_dict_by_path(self._cache, full_access_path)
        except (SerializationPathError, SerializationValueError, KeyError):
            return {}



import logging
from collections.abc import Callable
from copy import deepcopy
from typing import Any

from pydase.data_service.state_manager import StateManager
from pydase.observer_pattern.observable.observable_object import ObservableObject
from pydase.observer_pattern.observer.property_observer import (
    PropertyObserver,
)
from pydase.utils.helpers import get_object_attr_from_path_list
from pydase.utils.serializer import dump

logger = logging.getLogger(__name__)


class DataServiceObserver(PropertyObserver):
    def __init__(self, state_manager: StateManager) -> None:
        self.state_manager = state_manager
        self._notification_callbacks: list[
            Callable[[str, Any, dict[str, Any]], None]
        ] = []
        super().__init__(state_manager.service)

    def on_change(self, full_access_path: str, value: Any) -> None:
        cached_value_dict = deepcopy(
            self.state_manager._data_service_cache.get_value_dict_from_cache(
                full_access_path
            )
        )

        cached_value = cached_value_dict.get("value")
        if cached_value != dump(value)["value"] and all(
            part[0] != "_" for part in full_access_path.split(".")
        ):
            logger.debug("'%s' changed to '%s'", full_access_path, value)

        self._update_cache_value(full_access_path, value, cached_value_dict)

        for callback in self._notification_callbacks:
            callback(full_access_path, value, cached_value_dict)

        if isinstance(value, ObservableObject):
            self._update_property_deps_dict()

        self._notify_dependent_property_changes(full_access_path)

    def _update_cache_value(
        self, full_access_path: str, value: Any, cached_value_dict: dict[str, Any]
    ) -> None:
        value_dict = dump(value)
        if cached_value_dict != {}:
            if (
                cached_value_dict["type"] != "method"
                and cached_value_dict["type"] != value_dict["type"]
            ):
                logger.warning(
                    "Type of '%s' changed from '%s' to '%s'. This could have unwanted "
                    "side effects! Consider setting it to '%s' directly.",
                    full_access_path,
                    cached_value_dict["type"],
                    value_dict["type"],
                    cached_value_dict["type"],
                )
            self.state_manager._data_service_cache.update_cache(
                full_access_path,
                value,
            )

    def _notify_dependent_property_changes(self, changed_attr_path: str) -> None:
        changed_props = self.property_deps_dict.get(changed_attr_path, [])
        for prop in changed_props:
            # only notify about changing attribute if it is not currently being
            # "changed" e.g. when calling the getter of a property within another
            # property
            if prop not in self.changing_attributes:
                self._notify_changed(
                    prop,
                    get_object_attr_from_path_list(self.observable, prop.split(".")),
                )

    def add_notification_callback(
        self, callback: Callable[[str, Any, dict[str, Any]], None]
    ) -> None:
        """
        Registers a callback function to be invoked upon attribute changes in the
        observed object.

        This method allows for the addition of custom callback functions that will be
        executed whenever there is a change in the value of an observed attribute. The
        callback function is called with detailed information about the change, enabling
        external logic to respond to specific state changes within the observable
        object.

        Args:
            callback (Callable[[str, Any, dict[str, Any]]): The callback function to be
            registered. The function should have the following signature:
                - full_access_path (str): The full dot-notation access path of the
                  changed attribute. This path indicates the location of the changed
                  attribute within the observable object's structure.
                - value (Any): The new value of the changed attribute.
                - cached_value_dict (dict[str, Any]): A dictionary representing the
                  cached state of the attribute prior to the change. This can be useful
                  for understanding the nature of the change and for historical
                  comparison.
        """
        self._notification_callbacks.append(callback)



from enum import Enum


class ColouredEnum(Enum):
    """
    Represents a UI element that can display colour-coded text based on its value.

    This class extends the standard Enum but requires its values to be valid CSS
    colour codes. Supported colour formats include:
        - Hexadecimal colours
        - Hexadecimal colours with transparency
        - RGB colours
        - RGBA colours
        - HSL colours
        - HSLA colours
        - Predefined/Cross-browser colour names
    Refer to the this website for more details on colour formats:
    (https://www.w3schools.com/cssref/css_colours_legal.php)

    The behavior of this component in the UI depends on how it's defined in the data
    service:
        - As property with a setter or as attribute: Renders as a dropdown menu,
            allowing users to select and change its value from the frontend.
        - As property without a setter: Displays as a coloured box with the key of the
            `ColouredEnum` as text inside, serving as a visual indicator without user
            interaction.

    Example:
    --------
    ```python
    import pydase.components as pyc
    import pydase

    class MyStatus(pyc.ColouredEnum):
        PENDING = "#FFA500"  # Orange
        RUNNING = "#0000FF80"  # Transparent Blue
        PAUSED = "rgb(169, 169, 169)"  # Dark Gray
        RETRYING = "rgba(255, 255, 0, 0.3)"  # Transparent Yellow
        COMPLETED = "hsl(120, 100%, 50%)"  # Green
        FAILED = "hsla(0, 100%, 50%, 0.7)"  # Transparent Red
        CANCELLED = "SlateGray"  # Slate Gray

    class StatusExample(pydase.DataService):
        _status = MyStatus.RUNNING

        @property
        def status(self) -> MyStatus:
            return self._status

        @status.setter
        def status(self, value: MyStatus) -> None:
            # Custom logic here...
            self._status = value

    # Example usage:
    my_service = StatusExample()
    my_service.status = MyStatus.FAILED
    ```
    """



import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, ClassVar, SupportsIndex

if TYPE_CHECKING:
    from pydase.observer_pattern.observer.observer import Observer

logger = logging.getLogger(__name__)


class ObservableObject(ABC):
    _list_mapping: ClassVar[dict[int, "_ObservableList"]] = {}
    _dict_mapping: ClassVar[dict[int, "_ObservableDict"]] = {}

    def __init__(self) -> None:
        if not hasattr(self, "_observers"):
            self._observers: dict[str, list["ObservableObject | Observer"]] = {}

    def add_observer(
        self, observer: "ObservableObject | Observer", attr_name: str = ""
    ) -> None:
        if attr_name not in self._observers:
            self._observers[attr_name] = []
        if observer not in self._observers[attr_name]:
            self._observers[attr_name].append(observer)

    def _remove_observer(self, observer: "ObservableObject", attribute: str) -> None:
        if attribute in self._observers:
            self._observers[attribute].remove(observer)

    @abstractmethod
    def _remove_observer_if_observable(self, name: str) -> None:
        """Removes the current object as an observer from an observable attribute.

        This method is called before an attribute of the observable object is
        changed. If the current value of the attribute is an instance of
        `ObservableObject`, this method removes the current object from its list
        of observers. This is a crucial step to avoid unwanted notifications from
        the old value of the attribute.
        """

    def _notify_changed(self, changed_attribute: str, value: Any) -> None:
        """Notifies all observers about changes to an attribute.

        This method iterates through all observers registered for the object and
        invokes their notification method. It is called whenever an attribute of
        the observable object is changed.

        Args:
            changed_attribute (str): The name of the changed attribute.
            value (Any): The value that the attribute was set to.
        """
        for attr_name, observer_list in self._observers.items():
            for observer in observer_list:
                extendend_attr_path = self._construct_extended_attr_path(
                    attr_name, changed_attribute
                )
                observer._notify_changed(extendend_attr_path, value)

    def _notify_change_start(self, changing_attribute: str) -> None:
        """Notify observers that an attribute or item change process has started.

        This method is called at the start of the process of modifying an attribute in
        the observed `Observable` object. It registers the attribute as currently
        undergoing a change. This registration helps in managing and tracking changes as
        they occur, especially in scenarios where the order of changes or their state
        during the transition is significant.

        Args:
            changing_attribute (str): The name of the attribute that is starting to
            change. This is typically the full access path of the attribute in the
            `Observable`.
            value (Any): The value that the attribute is being set to.
        """

        for attr_name, observer_list in self._observers.items():
            for observer in observer_list:
                extended_attr_path = self._construct_extended_attr_path(
                    attr_name, changing_attribute
                )
                observer._notify_change_start(extended_attr_path)

    def _initialise_new_objects(self, attr_name_or_key: Any, value: Any) -> Any:
        new_value = value
        if isinstance(value, list):
            if id(value) in self._list_mapping:
                # If the list `value` was already referenced somewhere else
                new_value = self._list_mapping[id(value)]
            else:
                # convert the builtin list into a ObservableList
                new_value = _ObservableList(original_list=value)
                self._list_mapping[id(value)] = new_value
        elif isinstance(value, dict):
            if id(value) in self._dict_mapping:
                # If the list `value` was already referenced somewhere else
                new_value = self._dict_mapping[id(value)]
            else:
                # convert the builtin list into a ObservableList
                new_value = _ObservableDict(original_dict=value)
                self._dict_mapping[id(value)] = new_value
        if isinstance(new_value, ObservableObject):
            new_value.add_observer(self, str(attr_name_or_key))
        return new_value

    @abstractmethod
    def _construct_extended_attr_path(
        self, observer_attr_name: str, instance_attr_name: str
    ) -> str:
        """
        Constructs the extended attribute path for notification purposes, which is used
        in the observer pattern to specify the full path of an observed attribute.

        This abstract method is implemented by the classes inheriting from
        `ObservableObject`.

        Args:
            observer_attr_name (str): The name of the attribute in the observer that
            holds a reference to the instance. Equals `""` if observer itself is of type
            `Observer`.
            instance_attr_name (str): The name of the attribute within the instance that
            has changed.

        Returns:
            str: The constructed extended attribute path.
        """


class _ObservableList(ObservableObject, list[Any]):
    def __init__(
        self,
        original_list: list[Any],
    ) -> None:
        self._original_list = original_list
        ObservableObject.__init__(self)
        list.__init__(self, self._original_list)
        for i, item in enumerate(self._original_list):
            super().__setitem__(i, self._initialise_new_objects(f"[{i}]", item))

    def __setitem__(self, key: int, value: Any) -> None:  # type: ignore[override]
        if hasattr(self, "_observers"):
            self._remove_observer_if_observable(f"[{key}]")
            value = self._initialise_new_objects(f"[{key}]", value)
            self._notify_change_start(f"[{key}]")

        super().__setitem__(key, value)

        self._notify_changed(f"[{key}]", value)

    def append(self, __object: Any) -> None:
        self._initialise_new_objects(f"[{len(self)}]", __object)
        super().append(__object)
        self._notify_changed("", self)

    def clear(self) -> None:
        self._remove_self_from_observables()

        super().clear()

        self._notify_changed("", self)

    def extend(self, __iterable: Iterable[Any]) -> None:
        self._remove_self_from_observables()

        try:
            super().extend(__iterable)
        finally:
            for i, item in enumerate(self):
                super().__setitem__(i, self._initialise_new_objects(f"[{i}]", item))

            self._notify_changed("", self)

    def insert(self, __index: SupportsIndex, __object: Any) -> None:
        self._remove_self_from_observables()

        try:
            super().insert(__index, __object)
        finally:
            for i, item in enumerate(self):
                super().__setitem__(i, self._initialise_new_objects(f"[{i}]", item))

            self._notify_changed("", self)

    def pop(self, __index: SupportsIndex = -1) -> Any:
        self._remove_self_from_observables()

        try:
            popped_item = super().pop(__index)
        finally:
            for i, item in enumerate(self):
                super().__setitem__(i, self._initialise_new_objects(f"[{i}]", item))

            self._notify_changed("", self)
        return popped_item

    def remove(self, __value: Any) -> None:
        self._remove_self_from_observables()

        try:
            super().remove(__value)
        finally:
            for i, item in enumerate(self):
                super().__setitem__(i, self._initialise_new_objects(f"[{i}]", item))

            self._notify_changed("", self)

    def _remove_self_from_observables(self) -> None:
        for i in range(len(self)):
            self._remove_observer_if_observable(f"[{i}]")

    def _remove_observer_if_observable(self, name: str) -> None:
        key = int(name[1:-1])
        current_value = self.__getitem__(key)

        if isinstance(current_value, ObservableObject):
            current_value._remove_observer(self, name)

    def _construct_extended_attr_path(
        self, observer_attr_name: str, instance_attr_name: str
    ) -> str:
        if observer_attr_name != "":
            return f"{observer_attr_name}{instance_attr_name}"
        return instance_attr_name


class _ObservableDict(dict[str, Any], ObservableObject):
    def __init__(
        self,
        original_dict: dict[str, Any],
    ) -> None:
        self._original_dict = original_dict
        ObservableObject.__init__(self)
        dict.__init__(self)
        for key, value in self._original_dict.items():
            super().__setitem__(key, self._initialise_new_objects(f"['{key}']", value))

    def __setitem__(self, key: str, value: Any) -> None:
        if not isinstance(key, str):
            logger.warning("Converting non-string dictionary key %s to string.", key)
            key = str(key)

        if hasattr(self, "_observers"):
            self._remove_observer_if_observable(f"['{key}']")
            value = self._initialise_new_objects(key, value)
            self._notify_change_start(f"['{key}']")

        super().__setitem__(key, value)

        self._notify_changed(f"['{key}']", value)

    def _remove_observer_if_observable(self, name: str) -> None:
        key = name[2:-2]
        current_value = self.get(key, None)

        if isinstance(current_value, ObservableObject):
            current_value._remove_observer(self, name)

    def _construct_extended_attr_path(
        self, observer_attr_name: str, instance_attr_name: str
    ) -> str:
        if observer_attr_name != "":
            return f"{observer_attr_name}{instance_attr_name}"
        return instance_attr_name



import inspect
import logging
import warnings
from enum import Enum
from typing import TYPE_CHECKING, Any, get_type_hints

import rpyc  # type: ignore[import-untyped]

import pydase.units as u
from pydase.data_service.abstract_data_service import AbstractDataService
from pydase.data_service.task_manager import TaskManager
from pydase.observer_pattern.observable.observable import (
    Observable,
)
from pydase.utils.helpers import (
    convert_arguments_to_hinted_types,
    get_class_and_instance_attributes,
    get_object_attr_from_path_list,
    is_property_attribute,
    parse_list_attr_and_index,
    update_value_if_changed,
)
from pydase.utils.serializer import (
    Serializer,
    generate_serialized_data_paths,
    get_nested_dict_by_path,
)

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


def process_callable_attribute(attr: Any, args: dict[str, Any]) -> Any:
    converted_args_or_error_msg = convert_arguments_to_hinted_types(
        args, get_type_hints(attr)
    )
    return (
        attr(**converted_args_or_error_msg)
        if not isinstance(converted_args_or_error_msg, str)
        else converted_args_or_error_msg
    )


class DataService(rpyc.Service, AbstractDataService):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self._task_manager = TaskManager(self)

        if not hasattr(self, "_autostart_tasks"):
            self._autostart_tasks = {}

        filename = kwargs.pop("filename", None)
        if filename is not None:
            warnings.warn(
                "The 'filename' argument is deprecated and will be removed in a future "
                "version. Please pass the 'filename' argument to `pydase.Server`.",
                DeprecationWarning,
                stacklevel=2,
            )
            self._filename: str | Path = filename

        self.__check_instance_classes()
        self._initialised = True

    def __setattr__(self, __name: str, __value: Any) -> None:
        # Check and warn for unexpected type changes in attributes
        self._warn_on_type_change(__name, __value)

        # every class defined by the user should inherit from DataService if it is
        # assigned to a public attribute
        if not __name.startswith("_") and not inspect.isfunction(__value):
            self.__warn_if_not_observable(__value)

        # Set the attribute
        super().__setattr__(__name, __value)

    def _warn_on_type_change(self, attr_name: str, new_value: Any) -> None:
        if is_property_attribute(self, attr_name):
            return

        current_value = getattr(self, attr_name, None)
        if self._is_unexpected_type_change(current_value, new_value):
            logger.warning(
                "Type of '%s' changed from '%s' to '%s'. This may have unwanted "
                "side effects! Consider setting it to '%s' directly.",
                attr_name,
                type(current_value).__name__,
                type(new_value).__name__,
                type(current_value).__name__,
            )

    def _is_unexpected_type_change(self, current_value: Any, new_value: Any) -> bool:
        return (
            isinstance(current_value, float)
            and not isinstance(new_value, float)
            or (
                isinstance(current_value, u.Quantity)
                and not isinstance(new_value, u.Quantity)
            )
        )

    def __warn_if_not_observable(self, __value: Any) -> None:
        value_class = __value if inspect.isclass(__value) else __value.__class__

        if not issubclass(
            value_class,
            (int | float | bool | str | list | Enum | u.Quantity | Observable),
        ):
            logger.warning(
                "Class '%s' does not inherit from DataService. This may lead to"
                " unexpected behaviour!",
                value_class.__name__,
            )

    def __check_instance_classes(self) -> None:
        for attr_name, attr_value in get_class_and_instance_attributes(self).items():
            # every class defined by the user should inherit from DataService if it is
            # assigned to a public attribute
            if (
                not attr_name.startswith("_")
                and not inspect.isfunction(attr_value)
                and not isinstance(attr_value, property)
            ):
                self.__warn_if_not_observable(attr_value)

    def __set_attribute_based_on_type(  # noqa: PLR0913
        self,
        target_obj: Any,
        attr_name: str,
        attr: Any,
        value: Any,
        index: int | None,
        path_list: list[str],
    ) -> None:
        if isinstance(attr, Enum):
            update_value_if_changed(target_obj, attr_name, attr.__class__[value])
        elif isinstance(attr, list) and index is not None:
            update_value_if_changed(attr, index, value)
        elif isinstance(attr, DataService) and isinstance(value, dict):
            for key, v in value.items():
                self.update_DataService_attribute([*path_list, attr_name], key, v)
        elif callable(attr):
            process_callable_attribute(attr, value["args"])
        else:
            update_value_if_changed(target_obj, attr_name, value)

    def _rpyc_getattr(self, name: str) -> Any:
        if name.startswith("_"):
            # disallow special and private attributes
            raise AttributeError("cannot access private/special names")
        # allow all other attributes
        return getattr(self, name)

    def _rpyc_setattr(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            # disallow special and private attributes
            raise AttributeError("cannot access private/special names")

        # check if the attribute has a setter method
        attr = getattr(self, name, None)
        if isinstance(attr, property) and attr.fset is None:
            raise AttributeError(f"{name} attribute does not have a setter method")

        # allow all other attributes
        setattr(self, name, value)

    def write_to_file(self) -> None:
        """
        Serialize the DataService instance and write it to a JSON file.

        This method is deprecated and will be removed in a future version.
        Service persistence is handled by `pydase.Server` now, instead.
        """

        warnings.warn(
            "'write_to_file' is deprecated and will be removed in a future version. "
            "Service persistence is handled by `pydase.Server` now, instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if hasattr(self, "_state_manager"):
            self._state_manager.save_state()

    def load_DataService_from_JSON(  # noqa: N802
        self, json_dict: dict[str, Any]
    ) -> None:
        warnings.warn(
            "'load_DataService_from_JSON' is deprecated and will be removed in a "
            "future version. "
            "Service persistence is handled by `pydase.Server` now, instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        # Traverse the serialized representation and set the attributes of the class
        serialized_class = self.serialize()
        for path in generate_serialized_data_paths(json_dict):
            nested_json_dict = get_nested_dict_by_path(json_dict, path)
            value = nested_json_dict["value"]
            value_type = nested_json_dict["type"]

            nested_class_dict = get_nested_dict_by_path(serialized_class, path)
            class_value_type = nested_class_dict.get("type", None)
            if class_value_type == value_type:
                class_attr_is_read_only = nested_class_dict["readonly"]
                if class_attr_is_read_only:
                    logger.debug(
                        "Attribute '%s' is read-only. Ignoring value from JSON "
                        "file...",
                        path,
                    )
                    continue
                # Split the path into parts
                parts = path.split(".")
                attr_name = parts[-1]

                # Convert dictionary into Quantity
                if class_value_type == "Quantity":
                    value = u.convert_to_quantity(value)

                self.update_DataService_attribute(parts[:-1], attr_name, value)
            else:
                logger.info(
                    "Attribute type of '%s' changed from '%s' to "
                    "'%s'. Ignoring value from JSON file...",
                    path,
                    value_type,
                    class_value_type,
                )

    def serialize(self) -> dict[str, dict[str, Any]]:
        """
        Serializes the instance into a dictionary, preserving the structure of the
        instance.

        For each attribute, method, and property, the method includes its name, type,
        value, readonly status, and documentation if any in the resulting dictionary.
        Attributes and methods starting with an underscore are ignored.

        For nested DataService instances, the method serializes recursively.
        For attributes of type list, each item in the list is serialized individually.
        If an item in the list is an instance of DataService, it is serialized
        recursively.

        Returns:
            dict: The serialized instance.
        """
        return Serializer.serialize_object(self)["value"]

    def update_DataService_attribute(  # noqa: N802
        self,
        path_list: list[str],
        attr_name: str,
        value: Any,
    ) -> None:
        warnings.warn(
            "'update_DataService_attribute' is deprecated and will be removed in a "
            "future version. "
            "Service state management is handled by `pydase.data_service.state_manager`"
            "now, instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        # If attr_name corresponds to a list entry, extract the attr_name and the index
        attr_name, index = parse_list_attr_and_index(attr_name)
        # Traverse the object according to the path parts
        target_obj = get_object_attr_from_path_list(self, path_list)

        # If the attribute is a property, change it using the setter without getting the
        # property value (would otherwise be bad for expensive getter methods)
        if is_property_attribute(target_obj, attr_name):
            setattr(target_obj, attr_name, value)
            return

        attr = get_object_attr_from_path_list(target_obj, [attr_name])
        if attr is None:
            return

        self.__set_attribute_based_on_type(
            target_obj, attr_name, attr, value, index, path_list
        )
